/**
 * 
 */
package org.jboss.test.ws.core.soap;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.soap.SOAPFault;
import javax.xml.ws.soap.SOAPFaultException;

import org.jboss.ws.core.jaxws.SOAPFaultHelperJAXWS;
import org.jboss.ws.core.soap.SOAPMessageImpl;
import org.jboss.ws.core.soap.SOAPThrottler;
import org.jboss.ws.core.soap.ThrottlingConst;

import junit.framework.TestCase;

/**
 * @author amahmood
 *
 */
public class SOAPThrottlerTest extends TestCase
{

	//private Map metadata=null;
	//private InputStream inputStream=null;
	SOAPThrottler throttler=null;
	/**
	 * @param name
	 */
	public SOAPThrottlerTest(String name)
	{
		super(name);
	}

	/* (non-Javadoc)
	 * @see junit.framework.TestCase#setUp()
	 */
	protected void setUp() throws Exception
	{
		super.setUp();
		SOAPThrottler throttler = new SOAPThrottler();
		System.setProperty("javax.xml.soap.SOAPFactory", "org.jboss.ws.core.soap.SOAPFactoryImpl");
		System.setProperty("javax.xml.soap.MessageFactory", "org.jboss.ws.core.soap.MessageFactoryImpl");
	}

	/* (non-Javadoc)
	 * @see junit.framework.TestCase#tearDown()
	 */
	protected void tearDown() throws Exception
	{
		super.tearDown();
	}
	
	
	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#throttleRespose(java.util.Map, java.io.InputStream)}.
	 */
	public final void testThrottleRespose()
	{
		Map metadata = getMetadata(1024, 2048);
		
		assertNotNull(metadata);
		InputStream inputStream = getInputStream();
		assertNotNull(inputStream);

		throttler = new SOAPThrottler();
   	    InputStream newStream = throttler.throttleRespose(metadata, null);
   	    assertNotNull(newStream);
	}
	
	private Map getMetadata(int responseLimt, int contentLength)
	{
		Map metadata = new HashMap();
		Map header = new HashMap();
		header.put(ThrottlingConst.RESPONSE_LIMIT, Integer.toString(responseLimt));
		metadata.put("HEADER", header);
		List<String> list = new ArrayList<String>();
		list.add(Integer.toString(contentLength));
		metadata.put(ThrottlingConst.CONTENT_LENGTH,list);
		metadata.put(ThrottlingConst.CONTENT_RANGE,"1-2/"+Integer.toString(contentLength));
		metadata.put(ThrottlingConst.OVER_LIMIT_RESPONSE_ACTION, ThrottlingConst.DISCARD_RESPONSE);
		
		return metadata;
	}
	
	
	private InputStream getInputStream()
	{
		
			InputStream out = null;
			try
			{
				SOAPFault soapFault = javax.xml.soap.SOAPFactory.newInstance().createFault(ThrottlingConst.DISCARD_ERROR_MSG,
						org.jboss.ws.Constants.SOAP11_FAULT_CODE_CLIENT);
				soapFault.normalize();
				SOAPFaultException fe = SOAPFaultHelperJAXWS.getSOAPFaultException(soapFault);
				SOAPMessageImpl soapMessage = SOAPFaultHelperJAXWS.exceptionToFaultMessage(fe);

				
				ByteArrayOutputStream soapOutput = new ByteArrayOutputStream();
				soapMessage.writeTo(soapOutput);

				out = new ByteArrayInputStream(soapOutput.toByteArray());
			}
			catch (Exception e)
			{
				e.printStackTrace();
				out = null;
			}

		return out;
	}

	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#throttleRespose(java.util.Map, java.io.InputStream)}.
	 */
	public final void testThrottleRespose_NoResponseLimit()
	{
		Map metadata = getMetadata(0, 2048);
		metadata.remove("HEADER");
		assertNotNull(metadata);

		throttler = new SOAPThrottler();
   	    InputStream newStream = throttler.throttleRespose(metadata, null);
   	    assertNull(newStream);
	}
	
	
	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#throttleRespose(java.util.Map, java.io.InputStream)}.
	 */
	public final void testThrottleRespose_ContentLenghtOnly()
	{
		Map metadata = getMetadata(1024, 0);
		metadata.remove(ThrottlingConst.CONTENT_RANGE);
		assertNotNull(metadata);

		throttler = new SOAPThrottler();
   	    InputStream newStream = throttler.throttleRespose(metadata, null);
   	    assertNull(newStream);
	}
	
	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#throttleRespose(java.util.Map, java.io.InputStream)}.
	 */
	public final void testThrottleRespose_ContentRangeOnly()
	{
		Map metadata = getMetadata(1024, 0);
		metadata.remove(ThrottlingConst.CONTENT_LENGTH);
		assertNotNull(metadata);

		throttler = new SOAPThrottler();
   	    InputStream newStream = throttler.throttleRespose(metadata, null);
   	    assertNull(newStream);
	}
	
	
	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#throttleRespose(java.util.Map, java.io.InputStream)}.
	 */
	public final void testThrottleRespose_ContentLengthAndContentRange()
	{
		Map metadata = getMetadata(1024, 2048);
		
		assertNotNull(metadata);

		throttler = new SOAPThrottler();
   	    InputStream newStream = throttler.throttleRespose(metadata, null);
   	    assertNotNull(newStream);
	}
	
	
	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#throttleRespose(java.util.Map, java.io.InputStream)}.
	 */
	public final void testThrottleRespose_NoContentLengthAndContentRange()
	{
		Map metadata = getMetadata(100, 0);
		metadata.remove(ThrottlingConst.CONTENT_LENGTH);
		metadata.remove(ThrottlingConst.CONTENT_RANGE);
		assertNotNull(metadata);
		InputStream inputStream = getInputStream();
		assertNotNull(inputStream);

		throttler = new SOAPThrottler();
   	    InputStream newStream = throttler.throttleRespose(metadata, inputStream);
   	    assertNotNull(newStream);	
	}

	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#getResponseLimit(java.util.Map)}.
	 */
	public final void testGetResponseLimit()
	{
		
		Map metadata = new HashMap();
		Map header = new HashMap();
		header.put(ThrottlingConst.RESPONSE_LIMIT, "1024");
		metadata.put("HEADER", header);
		int value = SOAPThrottler.getResponseLimit(metadata);		
		assertEquals(1024, value);
	}

	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#getContentLength(java.util.Map)}.
	 */
	public final void testGetContentLength()
	{
		
		Map metadata = new HashMap();
		List<String> list = new ArrayList<String>();
		list.add("2048");
		metadata.put(ThrottlingConst.CONTENT_LENGTH,list);
		
		int value = SOAPThrottler.getContentLength(metadata);		
		assertEquals(2048, value);
	}

	/**
	 * Test method for {@link org.jboss.ws.core.soap.SOAPThrottler#getContentRange(java.util.Map)}.
	 */
	public final void testGetContentRange()
	{
		Map metadata = new HashMap();
		metadata.put(ThrottlingConst.CONTENT_RANGE,"1024-2048/4096");
		
		int value = SOAPThrottler.getContentRange(metadata);		
		assertEquals(4096, value);
	}

}
