/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the JBPM BPEL PUBLIC LICENSE AGREEMENT as
 * published by JBoss Inc.; either version 1.0 of the License, or
 * (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
package org.jbpm.bpel.integration.client;

import static org.jbpm.bpel.integration.client.ThrottlingConst.DISCARD_RESPONSE;
import static org.jbpm.bpel.integration.client.ThrottlingConst.OVER_LIMIT_RESPONSE_ACTION;
import static org.jbpm.bpel.integration.client.ThrottlingConst.RESPONSE_DISCARDED;
import static org.jbpm.bpel.integration.client.ThrottlingConst.RESPONSE_LIMIT;

import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.io.UnsupportedEncodingException;

import javax.wsdl.Fault;
import javax.wsdl.Part;
import javax.wsdl.Port;
import javax.wsdl.extensions.soap.SOAPAddress;
import javax.xml.namespace.QName;
import javax.xml.soap.MessageFactory;
import javax.xml.soap.MimeHeaders;
import javax.xml.soap.SOAPConnection;
import javax.xml.soap.SOAPConnectionFactory;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPEnvelope;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPFactory;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.commons.lang.builder.ToStringBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.MDC;
import org.jbpm.bpel.BpelException;
import org.jbpm.bpel.graph.exe.BpelFaultException;
import org.jbpm.bpel.graph.exe.FaultInstance;
import org.jbpm.bpel.integration.soap.MessageDirection;
import org.jbpm.bpel.integration.soap.SoapUtil;
import org.jbpm.bpel.integration.soap.SoapFormatter;
import org.jbpm.bpel.integration.soap.SoapUtil;
import org.jbpm.bpel.variable.def.MessageType;
import org.jbpm.bpel.variable.exe.MessageValue;
import org.jbpm.bpel.wsdl.xml.WsdlUtil;
import org.jbpm.bpel.xml.BpelConstants;
import org.jbpm.bpel.xml.schema.SchemaCacheManager;
import org.jbpm.bpel.xml.util.XmlUtil;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import sun.misc.BASE64Encoder;

import com.ibm.wsdl.extensions.soap.SOAPConstants;
import com..commonlogger.ContextTypeEnum;
import com..commonlogger.LogUtil;
import com..core.util.net.URLUtil;
import com..oe.OEContext;
import com..oe.branding.OEBrandingController;
/**
 * @author Alejandro Guzar
 * @version $Revision: 1.12 $ $Date: 2007/11/29 10:17:06 $
 */
public class SoapClient {

	private final SoapFormatter formatter;
	private URL address;
	private QName serviceName;
	
	//  additions
	private String authType = "SSO"; // default to SSO, for backward compatability
	private String simpleUserName;
	private String simplePassword;
        private Map partsMap;
        private SchemaCacheManager schemaCacheManager;

	private static final Log log = LogFactory.getLog(SoapClient.class);	
	
	public SoapClient(Port port) 
	{
		formatter = new SoapFormatter(port.getBinding());

		// exclude non-soap ports
		SOAPAddress soapAddress = (SOAPAddress) WsdlUtil.getExtension(port
				.getExtensibilityElements(), SOAPConstants.Q_ELEM_SOAP_ADDRESS);
		if (soapAddress == null)
			throw new BpelException("non-soap ports not supported: " + port);

		// exclude non-url address locations
		String location = soapAddress.getLocationURI();
		
		// additions
		if (location!= null) {
		    if (location.endsWith( "#SBMNONE")) {
		        location = location.substring( 0, location.lastIndexOf( "#"));
		        authType = "NONE";
		    }
		    if (location.endsWith( "#SBMSSO") ) {
		        location = location.substring( 0, location.lastIndexOf( "#"));
		        authType = "SSO";
		    } else if (location.endsWith( "#SBMSIMPLE")) {
		        int last = location.lastIndexOf( "#");
                int second = location.lastIndexOf( "#", last-1);
                int third = location.lastIndexOf( "#", second-1);
                simpleUserName = location.substring( third+1, second);
                simplePassword = location.substring( second+1, last);
                try {
                    simpleUserName = URLDecoder.decode( simpleUserName, "UTF-8");
                    simplePassword = URLDecoder.decode( simplePassword, "UTF-8");
                } catch (java.io.UnsupportedEncodingException ex) {
                    log.error("Cannot decode the url.");
                }
		        
                location = location.substring( 0, third);
                authType = "SIMPLE";		        
		    }
		}
		
		try 
		{
			address = new URL(URLUtil.encode(location, "UTF-8"));
		} 
		catch (MalformedURLException e) 
		{
			throw new BpelException("invalid address location: " + location, e);
		}
		catch (UnsupportedEncodingException uee) {
			throw new BpelException("UTF-8 encoding on URLs is not supported by this jdk " + uee);
		}		
	}

	public SoapFormatter getFormatter() {
		return formatter;
	}

	public URL getAddress() {
		return address;
	}

	public Map call(String operation, Map inputParts, Map globalInputParts,
			Map globalHeaderVariables, String stepName)
			throws BpelFaultException {
		try {
			SOAPMessage soapOutput = callImpl(operation, inputParts,
					globalInputParts, globalHeaderVariables, stepName);
			HashMap outputParts = new HashMap();

			if (SoapUtil.getElement(soapOutput.getSOAPBody(),
					javax.xml.soap.SOAPConstants.URI_NS_SOAP_ENVELOPE, "Fault") == null) {
				formatter.readMessage(operation, soapOutput, outputParts,
						MessageDirection.OUTPUT);
				return outputParts;
			}
	         // if outputParts above is not returned, it means that a fault happened
			
			if (isThrottlingFault(soapOutput))
			{
			    int throttlingLimit = ThrottlingProperties.getPayloadLimitInBytes();
				String message = LogUtil.createI18nLog("cannot.handle.payload", new String[]{stepName});
				MDC.put(ContextTypeEnum.MESSAGE_TYPE.toString(), "KEY");
				log.error(message);
				MDC.remove(ContextTypeEnum.MESSAGE_TYPE.toString());
				
				// Also log it to server.log. Thus this log cannot be translated.
                String messageForServerDotLog = "Execution of the orchestration workflow is stopped because the size of the Web service response exceeded the allowable limit.";
                ThrottlingFaultHelper.errorLogToFile( messageForServerDotLog);

                // Throw a named fault, which is to be caught in the generated BPEL.
                // This faultName shall be the same as the one used in BPEL_shell.txt in MC.
                QName faultQName = new QName("http://www.eclipse.org/alf/schema/EventBase/1", "ThrottlingFault");
                FaultInstance faultInstance = new FaultInstance(faultQName);
                throw new BpelFaultException(faultInstance);
			}
			
			String faultString = "";
			FaultInstance faultInstance = null;
			Fault fault = formatter.readFault(operation, soapOutput, outputParts);
			if (fault != null) {
				/*
				 * WS-BPEL 2.0 section 6.1: each WSDL fault is identified in
				 * WS-BPEL by a qualified name formed by the target namespace of
				 * the WSDL document in which the relevant port type and fault
				 * are defined, and the NCName of the fault
				 */
				String targetNamespace = formatter.getBinding().getPortType()
						.getQName().getNamespaceURI();
				QName faultQName = new QName(targetNamespace, fault.getName());

				MessageValue faultMessage = new MessageValue(new MessageType(
						fault.getMessage()));
				faultMessage.setParts(outputParts);

				faultInstance = new FaultInstance(faultQName,
						faultMessage);
				faultString = getPartsDetail (faultMessage);
				
			} else {
				// It means SoapFormatter did not find any wsdl fault matching
				// the soap fault.
				// Throw generic fault that can be handled with catchAll
				String targetNamespace = formatter.getBinding().getPortType()
						.getQName().getNamespaceURI();
				QName faultQName = new QName(targetNamespace, "fault");
				faultInstance = new FaultInstance(faultQName);
			}
	        MDC.put( ContextTypeEnum.MESSAGE_TYPE.toString(), "KEY");
	        String msg = LogUtil.createI18nLog( "handling.fault.exception", new String[] { stepName, 
	                                                                                       OEContext.getInstance().getBrandingController().getComposerProductNameFull() }, 
	                                                                                       faultString);
	        log.error(msg);
	        MDC.remove( ContextTypeEnum.MESSAGE_TYPE.toString());
	        throw new BpelFaultException(faultInstance);
	        
		} catch (SOAPException e) {
			e.printStackTrace();
			// BPEL-286 raise SOAP communication exception as BPEL fault
		    MDC.put( ContextTypeEnum.MESSAGE_TYPE.toString(), "KEY");
            String msg = LogUtil.createI18nLog("endpoint.call.failed", new String[]{stepName, 
                                                                                    address.toString(), 
                                                                                    OEContext.getInstance().getBrandingController().getOrchestrationProductName()});
			log.error(msg, e);
			MDC.remove( ContextTypeEnum.MESSAGE_TYPE.toString());
			throw new BpelFaultException(BpelConstants.FAULT_INVOCATION_FAILURE);
		}
	}
	
	private String getPartsDetail( MessageValue msg) {
        if ( msg == null)
            return null;
        Map parts = msg.getParts();

        if ( parts == null)
            return null;
        StringBuilder sb = new StringBuilder();
        for ( Object key : parts.keySet()) {
            //sb.append( key).append( " : ");
            try {
                Node node = (Node) parts.get( key);
                String str = SoapUtil.nodeToString( node);
                sb.append( str);
            }
            catch ( Exception e) {
            }
            sb.append( ". ");
        }
        return sb.toString();
    }

	public void callOneWay(String operation, Map inputParts,
			Map globalInputParts, Map globalHeaderVariables, String stepName) {
		try {
			callImpl(operation, inputParts, globalInputParts,
					globalHeaderVariables, stepName);
		} catch (SOAPException e) {
			// BPEL-286 raise SOAP communication exception as BPEL fault
            MDC.put( ContextTypeEnum.MESSAGE_TYPE.toString(), "KEY");
            String msg = LogUtil.createI18nLog( "endpoint.call.failed", new String[]{stepName, address.toString()});
            log.error(msg, e);
            MDC.remove( ContextTypeEnum.MESSAGE_TYPE.toString());
			throw new BpelFaultException(BpelConstants.FAULT_INVOCATION_FAILURE);
		}
	}

	protected SOAPMessage callImpl(String operation, Map inputParts,
			Map globalInputParts, Map globalHeaderVariables, String stepName)
			throws SOAPException {
		// create message
		MessageFactory messageFactory = MessageFactory.newInstance();
		SOAPMessage soapInput = messageFactory.createMessage();

		SOAPEnvelope soapEnv = soapInput.getSOAPPart().getEnvelope();
		soapEnv.addNamespaceDeclaration("xsd",
				"http://www.w3.org/2001/XMLSchema");
		SOAPHeader soapHeader = soapEnv.getHeader();
		//soapHeader.addChildElement("soap-max-response-size", "1024");
		MimeHeaders mimeHeaders = soapInput.getMimeHeaders();
        int throttlingLimit = ThrottlingProperties.getPayloadLimitInBytes();

		if (throttlingLimit > 0)
		{
			mimeHeaders.setHeader(RESPONSE_LIMIT, Integer.toString(throttlingLimit));
			mimeHeaders.setHeader(OVER_LIMIT_RESPONSE_ACTION, DISCARD_RESPONSE);
		}
		SOAPFactory soapFactory = SOAPFactory.newInstance();

		// Read wsse:security element from globalInputParts
		Element securityElement = null;
		Iterator iterator = null;
		if( globalInputParts != null )
		{
			iterator = globalInputParts.keySet().iterator();
			if (iterator.hasNext()) {
				Object key = iterator.next();
				Element inputElement = (Element) globalInputParts.get(key);
				// WSSE Security element is only added if SSO is the authentication
				// type of the endpoint defined.
				if ("SSO".equals( authType) ) {
    				securityElement = getWsseSecurityElement(inputElement);
    				if (securityElement != null) {
    					// add security element to outgoing header
    						createSoapElements(soapHeader, securityElement,
    								soapFactory, operation);
    				}
				}
			} else {
				throw new SOAPException(
						" BPEL process cannot accept input with no parts");
			}
		}

		// Copy SOAP and HTTP header variables elements
		if( globalHeaderVariables != null )
		{
			iterator = globalHeaderVariables.keySet().iterator();
			while (iterator.hasNext()) {
				Element headerElement = (Element) globalHeaderVariables
						.get(iterator.next());
				createSoapElements(soapHeader, headerElement, soapFactory, operation);
				createHttpHeader(mimeHeaders, headerElement);
			}
		}

        //DEF172542  FF: Orchestration: ws data elements ordering mismatch for complex types 
 		//Before sending the SOAP request to external webservice, validate the request.
		HashMap newInputParts = null;
		if( partsMap != null && schemaCacheManager != null) {
		    newInputParts = SoapUtil.validateSoapMessage(inputParts, partsMap, schemaCacheManager );
		    // write message
			formatter.writeMessage(operation, soapInput, newInputParts,
					MessageDirection.INPUT);
		}
		else
		{
			// write message
			formatter.writeMessage(operation, soapInput, inputParts,
					MessageDirection.INPUT);
		}
		// call endpoint
		SOAPConnection soapConnection = SOAPConnectionFactory.newInstance()
				.createConnection();
		try {
		    // If the endpoint has simple/basic authentication setting, uses it.
		    // Otherwise, use the setting defined in the workflow.
		    if ("SIMPLE".equals( authType)) {
		        String authorization = new BASE64Encoder().encode((simpleUserName
                    + ":" + simplePassword).getBytes());
		        //DEF181792 http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6459815
		        //This is not related to non-ascii chars as the above defect says.
		        //If encoded string goes above 76, Encoder class adds \n throwing illegal character exception.
		        authorization = authorization.replaceAll("\r\n?", "");
		        mimeHeaders.addHeader("Authorization", "Basic "
                    + authorization);
		    }
		    else if( globalHeaderVariables != null )
			{
		        
		        
				Element usernameE = (Element) globalHeaderVariables
						.get(stepName + "-BAA_UserName");
				if (usernameE != null) {
					String username = usernameE.getTextContent();
					Element passwordE = (Element) globalHeaderVariables
							.get(stepName + "-BAA_Password");
					String password;
	
					if (passwordE != null)
						password = passwordE.getTextContent();
					else
						password = "";
	
					String authorization = new BASE64Encoder().encode((username
							+ ":" + password).getBytes());
					//DEF181792 http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6459815
	                //This is not related to non-ascii chars as the above defect says.
	                //If encoded string goes above 76, Encoder class adds \n throwing illegal character exception.
					authorization = authorization.replaceAll("\r\n?", "");
						mimeHeaders.addHeader("Authorization", "Basic "
								+ authorization);
				}
			}

			// SAML token may appear twice, once in the header and once in the
			// body so 2 is passed in.
            SoapUtil.logSOAPContent("SOAP.message.sent.out", new String[]{stepName,
                                          OEContext.getInstance().getBrandingController().getOrchestrationProductName()}, soapInput, 2);

            SOAPMessage soapOutput = soapConnection.call(soapInput,
					getAddress());
            SoapUtil.logSOAPContent("SOAP.message.received", new String[]{stepName,
                                                                                OEContext.getInstance().getBrandingController().getOrchestrationProductName()}, soapOutput, 0);
			return soapOutput;
		} finally {
			soapConnection.close();
		}
	}

	private void createHttpHeader(MimeHeaders mimeHeaders, Element domElement) {
		String newVarName = null;
		String headerName = domElement.getLocalName();
		String headerValue = null;
		if (headerName.endsWith("HttpHeader")) {
			// remove the suffix
			try {
				newVarName = headerName.substring(0, headerName
						.lastIndexOf("HttpHeader"));
				headerValue = domElement.getTextContent();
			} catch (Exception e) {
			}
		} else {
			// This is not Http header so return
			return;
		}
		if (newVarName != null && newVarName.length() > 0
				&& headerValue != null)
			mimeHeaders.addHeader(newVarName, headerValue);
	}

	public void createSoapElements(SOAPElement parentSoapElement,
			Element domElement, SOAPFactory soapFactory, String operation)
			throws SOAPException {
		// Create placeholder SOAPElement from Element

		// TODO: THIS IS A HACK TO SEND CORRECT NAMESPACE FOR SALESFORCE HEADER
		// ELEMENTS
		/*
		 * String namespace = domElement.getNamespaceURI(); String name =
		 * domElement.getLocalName(); if( name.equals( "SessionHeader") ||
		 * name.equals( "sessionId")) { namespace =
		 * "urn:enterprise.soap.sforce.com"; }
		 */
		String newVarName = domElement.getLocalName();
		String headerName = domElement.getLocalName();
		//construct the soap header prefix based on convention <service name>_<operation name>. 
		//For backward compatibility, leave the prefix as stepName. This fixes defect DEF164654
		if(serviceName == null)
			throw new SOAPException("Service cannot be null while making SOAP call");
		
		String soapHeaderElementPrefix = serviceName.getLocalPart() + "_" + operation;
		
		//DEF168664 : Removed working data SoapHeader support which looked for SoapHeader suffix on variable name
		
		if (headerName.endsWith("HttpHeader")
				|| headerName.endsWith("-BAA_UserName")
				|| headerName.endsWith("-BAA_Password")) {
			// This is HttpHeader so return
		} else if (headerName.equals(soapHeaderElementPrefix + "_Envelope")) 
			// New Envelope element handling. We don't send Envelope but
		    // all its children as header if name matches
		{
			
			NodeList soapHeaderElements = domElement.getChildNodes();
			for (int i = 0, len = soapHeaderElements.getLength(); i < len; i++) {
				Node node = soapHeaderElements.item(i);
				if (node.getNodeType() != Node.ELEMENT_NODE)
					continue;
				createSoapElements(parentSoapElement, (Element) node,
						soapFactory, operation);
			}
			// Done adding children of Envelope. Return.
			
		}
		else
		{
			SOAPElement newSoapElement = null;
			if (newVarName != null && newVarName.length() > 0)
				newSoapElement = soapFactory.createElement(newVarName, domElement
						.getPrefix(), domElement.getNamespaceURI());
			else
				newSoapElement = soapFactory.createElement(domElement
						.getLocalName(), domElement.getPrefix(), domElement
						.getNamespaceURI());
	
			newSoapElement = parentSoapElement.addChildElement(newSoapElement);
	
			// attributes
		    SoapUtil.copyAttributes(newSoapElement, domElement);
		    // all namespaces
		    SoapUtil.copyVisibleNamespaces(newSoapElement, domElement);
		    XmlUtil.ensureOwnNamespaceDeclared(newSoapElement);
		    
			/*
			// Add any attribute on Element as attributes of SOAPElement
			NamedNodeMap attributes = domElement.getAttributes();
			int len = attributes.getLength();
			for (int i = 0; i < len; i++) {
				Attr attr = (Attr) attributes.item(i);
				if (attr.getLocalName() != null) {
					newSoapElement.addAttribute(soapFactory.createName(attr
							.getLocalName(), attr.getPrefix(), attr
							.getNamespaceURI()), attr.getValue());
				} else if (attr.getName() != null) {
					newSoapElement.addAttribute(soapFactory.createName(attr
							.getName()), attr.getValue());
				}
	
			}
			*/
			// Add children of this Element to SOAPElement. This will call
			// recurrsively
			NodeList nodeList = domElement.getChildNodes();
			int len = nodeList.getLength();
			for (int j = 0; j < len; j++) {
				Node childNode = nodeList.item(j);
				if (childNode instanceof org.w3c.dom.Text) {
					newSoapElement.addTextNode(childNode.getNodeValue());
				} else if (childNode instanceof Element) {
					createSoapElements(newSoapElement, (Element) childNode,
							soapFactory, operation);
				}
			}
		}
		return;
	}

	private Element getWsseSecurityElement(Element inputElement) {
		XPathFactory factory = XPathFactory.newInstance();
		XPath xpath = factory.newXPath();
		xpath.setNamespaceContext(new NamespaceContext());
		NodeList nodeList = null;
		try {
			XPathExpression expr = xpath.compile("//wsse:Security");
			nodeList = (NodeList) expr.evaluate(inputElement,
					XPathConstants.NODESET);
		} catch (XPathExpressionException e) {
			return null;
		}
		if (nodeList == null)
			return null;
		if (nodeList.item(0) == null)
			return null;
		return (Element) nodeList.item(0);
	}

	public String toString() {
		return new ToStringBuilder(this).append("formatter", formatter).append(
				"address", getAddress()).toString();
	}

	public void setAddressIfModified(String newAddress) {

		if (newAddress == null)
			return;
		
		URL newUrl = null;
		try {
			newUrl = new URL(URLUtil.encode(newAddress, "UTF-8"));
		} catch (MalformedURLException e) {
			throw new BpelException("invalid address location: " + newAddress,
					e);
		}
		catch (UnsupportedEncodingException uee) {
			throw new BpelException("UTF-8 encoding on URLs is not supported by this jdk " + uee);
		}
				
		if ( address.toString().equals(newUrl.toString()))
			return;

		address = newUrl;
	}
	
	/**
	 * @param soapOutput
	 * @return true if fault was due to throttled response.
	 */
	private boolean isThrottlingFault(SOAPMessage soapOutput)
	{
		boolean out = false;
		MimeHeaders mh = soapOutput.getMimeHeaders();
		String[] values = mh.getHeader(OVER_LIMIT_RESPONSE_ACTION);

		if (values != null && values.length > 0)
		{
			if (values[0].equals(RESPONSE_DISCARDED))
			{
				out = true;
				log.debug("Throttling SOAPFault found.");
			}
		}

		return out;
	}

	public void setService(QName name) {
		serviceName = name;
	}
	
	/**
	 * This method stores parts map for current operation.
	 * SoapClient uses the parts map to look up the schema type for each message part.
	 * @param partsMap
	 */
    public void setParts( Map partsMap) {
        this.partsMap = partsMap;        
    }

    public void setSchemaCacheManager( SchemaCacheManager schemaCacheManager) {
        this.schemaCacheManager = schemaCacheManager;
    }
	
}
